/*
   program : front

   $Source: /home/yee/C/RCS/front.c,v $
   $Author: yee $
   $Date: 1999/02/17 07:53:00 $

   $Revision: 1.36 $

   This program gives a curses based frontend to programs.

   Compile Options : none

   Runtime Parameters : <input filename> <output filename>

   Assumptions : none known

   Bugs :
     1) max string length read on entry restricted to QUANTUM (1024).

   To Do :
     1) support text entry of float restricted options.
     2) if possible, combine code for int and string restricted options.
     3) support options that do not have a value associated with it.
     4) support options that may or may not be used.

   $Log: front.c,v $
   Revision 1.36  1999/02/17 07:53:00  yee
   documentation change

   Revision 1.35  1999/02/16 16:06:35  yee
   working restricted space text entry for int and text.
   restricted space entry of floats not done yet.

   Revision 1.28  1999/02/15 06:14:41  yee
   Able to alter free entries as well as restricted space text entries.
   Modification of restricted space int and floats not yet working.

   Revision 1.19  1999/02/06 20:03:55  yee
   cleaned up logic.

   Revision 1.18  1999/02/06 19:44:07  yee
   Simplified StringLength functions.

   Revision 1.17  1999/02/06 17:49:48  yee
   changed to generic function, AddElement(), to handle double linked lists.

   Revision 1.13  1999/01/24 22:26:02  yee
   working.

   Copyright 1999 by Conway Yee.  All rights reserved.
 */
char front[] = "$Id: front.c,v 1.36 1999/02/17 07:53:00 yee Exp $";

#include <ctype.h>
#include <curses.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "style.h"
#include "util.h"
#include "mutil.h"

#define INITCURSES(soft) {  /* initialize curses */ \
          slk_init(soft);                           \
          initscr();                                \
          cbreak();                                 \
          noecho();                                 \
          nl();                                     \
          intrflush(stdscr,FALSE);                  \
          keypad(stdscr,TRUE);                      \
        } /* initialize curses */

#define wblankout(win,i) { /* blank out extra characters */ \
          int counter=i;                                    \
          for(;counter;counter--)                           \
             wprintw(win," ");                              \
        }

enum{NONE,FLOAT,INT,CHAR,STRING};

union value {
  float f;
  int i;
  char c;
  char *s;
} value;
                          /* Note: structures MUST have next and prev
                                   as the first members of the structures */
typedef struct val {
  struct val *next;        /* next value */
  struct val *prev;        /* previous value */
  union value v;
} val;

typedef struct option {
  struct option *next;     /* next line */
  struct option *prev;     /* previous line */
  char *text;              /* intro text */
  char *opt;               /* option code */
  int maxstr;              /* maximum string length */
  int y,x;                 /* cursor position after intro text written */
  CHOICE type;             /* data type */
  union value incr;        /* delta step when changing values */
  struct val *val;         /* value */
} option;

typedef struct window {
  struct window *next;     /* next window */
  struct window *prev;     /* previous window */
  char *title;             /* window title */
  char *exec;              /* execution string with fixed parameters */
  struct option *opt;      /* double linked list of options */
  WINDOW *win;             /* curses window # */
} window;

PRIVATE window *GetWindow(window *wptr,FILE *fptr);
PRIVATE void WriteScript(char *fname,window *w);
PRIVATE window *ReadDataFile(FILE *fptr);
PRIVATE void manipulate(char *fname,window *w);
PRIVATE void SetWindow(window *w);
PRIVATE void AlterWindow(char *fname,window *w);
PRIVATE void CloseWindow(window *w);

#define INCREMENT_QUANTUM 100.0  /* sets amount to change values */

#define OUTSUFFIX "sh"           /* file naming convention */
#define INSUFFIX "dat"
#define FNAME "front"
#define DELIM "."

#define QUANTUM 1024             /* max length of string read */
 
/*****************************************************************************/
/*****************************************************************************/

PUBLIC void main(int argc,char *argv[])

{
  window *w;
  FILE *infile;
  char *outfilename;

  switch(argc) {
  case 1:
    {
      char *str=malloc(sizeof(FNAME)+sizeof(DELIM)+sizeof(INSUFFIX)+1);

      strcpy(str,FNAME);  /* construct file name and open */
      strcat(str,DELIM);
      strcat(str,INSUFFIX);
      infile=fopen(str,"r");

      outfilename=realloc(str,sizeof(FNAME)+sizeof(DELIM)+sizeof(OUTSUFFIX)+1);
      strcpy(str,FNAME);  /* construct file name */
      strcat(str,DELIM);
      strcat(str,OUTSUFFIX);
    }
    break;
  case 2:
    infile=fopen(argv[1],"r");

    {
      char *temp;   /* location of DELIM */

      outfilename=malloc(sizeof(argv[1])+sizeof(DELIM)+sizeof(OUTSUFFIX)+1);
      strcpy(outfilename,argv[1]);
      temp=strpbrk(outfilename,DELIM);
      if(temp)
        strcpy(temp+1,OUTSUFFIX);
      else {
        strcat(outfilename,DELIM);
        strcat(outfilename,OUTSUFFIX);
      }
    }
    break;
  default: /* both infile and outfile specified */
    infile=fopen(argv[1],"r");

    outfilename=argv[2];
  }

  if(infile) {
    w=ReadDataFile(infile);

    manipulate(outfilename,w);

    WriteScript(outfilename,w);  /* save final format */
  } else
    fprintf(stderr,"front : unable to open infile\n");
}

/*****************************************************************************/

/* uses a curses interface to manipulate parameter values */

PRIVATE void manipulate
(
 char *fname,    /* output filename */
 window *w      /* data structure */
)
{
  fprintf(stdout,"PAGE DOWN KEY           next program\n");
  fprintf(stdout,"PAGE UP KEY             previous program\n");
  fprintf(stdout,"DOWN ARROW KEY          next option\n");
  fprintf(stdout,"UP ARROW KEY            previous option\n");
  fprintf(stdout,"RIGHT ARROW KEY         increment option\n");
  fprintf(stdout,"LEFT ARROW KEY          decrement option\n");
  fprintf(stdout,"F7 KEY                  execute\n");
  fprintf(stdout,"F8 KEY                  quit\n");
  fprintf(stdout,"\nPRESS ENTER KEY TO CONTINUE\n");
  getchar();

  INITCURSES(0);
  { /* set soft labels */
    slk_set(1,"PDN WIN+",0);
    slk_set(2,"PUP WIN-",0);
    slk_set(3,"DN  OPT+",0);
    slk_set(4,"UP  OPT-",0);
    slk_set(5,"RT  VAL+",0);
    slk_set(6,"LT  VAL-",0);
    
    slk_set(7,"F7  GO ",0);
    slk_set(8,"F8 QUIT",0);

    slk_touch();
    slk_refresh();
  } /* set soft labels */

  SetWindow(w);
  AlterWindow(fname,w);
  CloseWindow(w);

  endwin();
}

/*****************************************************************************/

/* initializes windows */

PRIVATE void SetWindow
(
 window *w      /* data structure */
)
{
  window *cwin=w;  /* current window */
  do {
                           /* only create window for structures with options */
    if(cwin->opt!=NULL) {
      option *copt=cwin->opt;          /* list of options */
      int wline=2;                     /* line number to write option */

      cwin->win=newwin(LINES-1,0,0,0); /* initialize window */
      keypad(cwin->win,TRUE);

      { /* writes title string */
        int size=strlen(cwin->title);  /* size of title string */

        wmove(cwin->win,0,(COLS>>1)-(size>>1));    /* center title string */
        wprintw(cwin->win,"%s\n",cwin->title);
        wmove(cwin->win,0,0);
      } /* writes title string */

      do {
        wmove(cwin->win,wline,0);
        wprintw(cwin->win,"%s : ",copt->text);
        getyx(cwin->win,copt->y,copt->x);

        switch(copt->type) {
        case INT:
          wprintw(cwin->win,"%d",copt->val->v.i);
          break;
        case FLOAT:
          wprintw(cwin->win,"%g",copt->val->v.f);
          break;
        case STRING:
          wprintw(cwin->win,"%s",copt->val->v.s);
          break;
        }

        wline++;
        copt=copt->next;
      } while(copt!=cwin->opt);

    }
    cwin=cwin->next;                 /* next window */
  } while(cwin!=w);

}

/*****************************************************************************/

/* this function manipulates the values of the various parameters */

PRIVATE void AlterWindow
(
 char *fname,   /* output filename */
 window *w      /* data structure */
)
{
  window *cwin=w;          /* current window */
  option *copt=w->opt;     /* current option */
  BOOLEAN done=FALSE;         /* terminate loop */

  touchwin(cwin->win);
  wrefresh(cwin->win);

  do {
    int ch;          /* character read */

    wmove(cwin->win,copt->y,copt->x);
    ch=wgetch(cwin->win);

    switch(ch) {
    case KEY_NPAGE:                  /* next window */
      do {
        cwin=cwin->next;
        copt=cwin->opt;
      } while(copt==NULL);
      touchwin(cwin->win);
      break;
    case KEY_PPAGE:                  /* previous window */
      do {
        cwin=cwin->prev;
        copt=cwin->opt;
      } while(copt==NULL);
      touchwin(cwin->win);
      break;
    case KEY_DOWN:                   /* next option */
      copt=copt->next; 
      break;
    case KEY_UP:                     /* previous option */
      copt=copt->prev;
      break;
    case KEY_RIGHT:                  /* next/increment value */
      if((copt->val!=copt->val->next)&&(copt->val!=copt->val->prev))
        copt->val=copt->val->next;
      else
        switch(copt->type) {
        case INT:
          if(copt->incr.i==0) {
            copt->incr.i=NINT(fabs(copt->val->v.i/INCREMENT_QUANTUM));
            if(copt->incr.i==0) copt->incr.i=1;
          }
          copt->val->v.i += copt->incr.i;
          break;
        case FLOAT:
          if(copt->incr.f==0.0)
            copt->incr.f = fabs(copt->val->v.f/INCREMENT_QUANTUM);
          copt->val->v.f += copt->incr.f;
          break;
        case STRING:
          flash();
          break;
        }
      switch(copt->type) {
      case INT:
        mvwprintw(cwin->win,copt->y,copt->x,"%d",copt->val->v.i);
        wclrtoeol(cwin->win);
        break;
      case FLOAT:
        mvwprintw(cwin->win,copt->y,copt->x,"%g",copt->val->v.f);
        wclrtoeol(cwin->win);
        break;
      case STRING:
        mvwprintw(cwin->win,copt->y,copt->x,"%s",copt->val->v.s);
        wclrtoeol(cwin->win);
        break;
      }
      break;
    case KEY_LEFT:                   /* previous/decrement value */
      if((copt->val!=copt->val->next)&&(copt->val!=copt->val->prev))
        copt->val=copt->val->prev;
      else
        switch(copt->type) {
        case INT:
          if(copt->incr.i==0) {
            copt->incr.i=NINT(fabs(copt->val->v.i/(float)INCREMENT_QUANTUM));
            if(copt->incr.i==0) copt->incr.i=1;
          }
          copt->val->v.i -= copt->incr.i;
          break;
        case FLOAT:
          if(copt->incr.f==0.0)
            copt->incr.f = fabs(copt->val->v.f/INCREMENT_QUANTUM);
          copt->val->v.f -= copt->incr.f;
          break;
        case STRING:
          flash();
          break;
        }
      switch(copt->type) {
      case INT:
        mvwprintw(cwin->win,copt->y,copt->x,"%d",copt->val->v.i);
        wclrtoeol(cwin->win);
        break;
      case FLOAT:
        mvwprintw(cwin->win,copt->y,copt->x,"%g",copt->val->v.f);
        wclrtoeol(cwin->win);
        break;
      case STRING:
        mvwprintw(cwin->win,copt->y,copt->x,"%s",copt->val->v.s);
        wclrtoeol(cwin->win);
        break;
      }
      break;
    case KEY_F(7):                   /* execute */
      WriteScript(fname,w);
      {
        WINDOW *execwin=newwin(LINES-1,0,0,0);
        
        wrefresh(execwin);
        system(fname);
        wclear(execwin);
        wrefresh(execwin);
        delwin(execwin);
      }
      break;
    case KEY_F(8):                   /* exit */
      done=TRUE;
      break;
    default:
      echo();
      ungetch(ch);
                                                              /* free entry */
      if((copt->val==copt->val->next)&&(copt->val==copt->val->prev)) {
        switch(copt->type) {
        case INT:
          wscanw(cwin->win,"%d",&(copt->val->v.i));
          mvwprintw(cwin->win,copt->y,copt->x,"%d",copt->val->v.i);
          wclrtoeol(cwin->win);
          break;
        case FLOAT:
          wscanw(cwin->win,"%g",&(copt->val->v.f));
          mvwprintw(cwin->win,copt->y,copt->x,"%g",copt->val->v.f);
          wclrtoeol(cwin->win);
          break;
        case STRING:
          {
            char *t=malloc(QUANTUM);           /* temporary string */
            wscanw(cwin->win,"%s",t);
            free(copt->val->v.s);
            t=realloc(t,sizeof(t));
            copt->val->v.s=t;
          }
          mvwprintw(cwin->win,copt->y,copt->x,"%s",copt->val->v.s);
          wclrtoeol(cwin->win);
          break;
        }
      } else {   /* restricted value space */
        switch(copt->type) {
        case INT: 
          {
            BOOLEAN LeadingZeros=FALSE;     /* are there leading zeros? */
            char *str1=malloc(QUANTUM);     /* dummy string */
            char *str2=malloc(QUANTUM);     /* dummy string */

            if(isspace(ch)) {
              do {                         /* eliminate initial white space */
                ch=wgetch(cwin->win);
              } while (isspace(ch)&&(ch!=NL));

              if(ch==ZERO) {
                LeadingZeros=TRUE;
                do {                         /* eliminate leading zero's */
                  ch=wgetch(cwin->win);
                } while (ch==ZERO);
              }

              if(isdigit(ch)||(ch==MINUS))     /* deal with first real digit */
                ungetch(ch);
            }

            if(!isspace(ch)) {
              val *bval=copt->val;  /* possible option with beginning OK */

              {
                int i=0;            /* counter */
                BOOLEAN done;      /* appropriate option found */

                for(done=FALSE;!done;) {
                  ch=wgetch(cwin->win);

                  if((ch==NL)||(ch==NUL))
                    done=TRUE;
                  else if (LeadingZeros&&(i==0)&&(ch==MINUS)) {
                    done=TRUE;
                    flash();
                  } else {
                    sprintf(str1,"%d",bval->v.i);
                    if((i==1)&&(str1[0]==MINUS)) {
                      ungetch(ch);               /* eliminate leading zeros */
                      do {
                        ch=wgetch(cwin->win);
                      } while(ch==ZERO);
                    }
                    if(ch==str1[i]) {           /* current option OK */
                      i++;
                      mvwprintw(cwin->win,copt->y,copt->x,"%d",bval->v.i);
                      wclrtoeol(cwin->win);
                      wmove(cwin->win,copt->y,copt->x+i);
                    } else {                          /* try other options */
                      BOOLEAN match=FALSE;  /* found matching value? */
                      val *tval;            /* option to test. bval no good */

                      tval=bval->next;
                      do {
                        BOOLEAN PreviousOK;     /* earlier characters OK? */

                        sprintf(str1,"%d",tval->v.i);
                        sprintf(str2,"%d",bval->v.i);

                        if(i>0)
                          PreviousOK = !strncmp(str1,str2,i-1);
                        else
                          PreviousOK = TRUE;
                        if((ch==str1[i])&&PreviousOK) {
                          match=TRUE;
                          i++;
                        } else
                          tval=tval->next;
                      } while((!match)&&(tval!=copt->val));
                      mvwprintw(cwin->win,copt->y,copt->x,"%d",tval->v.i);
                      wclrtoeol(cwin->win);
                      if(tval==copt->val) {     /* error */
                        done=TRUE;
                        flash();
                      } else {
                        bval=tval;              /* set current guess */
                        wmove(cwin->win,copt->y,copt->x+i);
                      }
                    }
                  }
                }
                copt->val=bval;
              }
            }
            mvwprintw(cwin->win,copt->y,copt->x,"%d",copt->val->v.i);
            wclrtoeol(cwin->win);
            wrefresh(cwin->win);
            free(str1);
            free(str2);
            break;
          }
        case FLOAT:
          wgetch(cwin->win);
          flash();
          break;
        case STRING:
          if(isspace(ch)) {
            do {                            /* eliminate initial white space */
              ch=wgetch(cwin->win);
            } while (isspace(ch)&&(ch!=NL));

            if(!isspace(ch))                /* first non-zero character */
              ungetch(ch);
          }

          if(!isspace(ch)) {
            val *bval=copt->val;  /* possible option with beginning OK */

            {
              int i=0;            /* counter */
              BOOLEAN done;      /* appropriate option found */

              for(done=FALSE;!done;) {
                ch=wgetch(cwin->win);

                if((ch==NL)||(ch==NUL))
                  done=TRUE;
                else if(ch==bval->v.s[i]) {           /* current option OK */
                  i++;
                  mvwprintw(cwin->win,copt->y,copt->x,"%s",bval->v.s);
                  wclrtoeol(cwin->win);
                  wmove(cwin->win,copt->y,copt->x+i);
                } else {                              /* try other options */
                  BOOLEAN match=FALSE;  /* found matching value? */
                  val *tval;            /* option to test. bval no good */

                  tval=bval->next;
                  do {
                    BOOLEAN PreviousOK;     /* earlier characters OK? */
                    if(i>0)
                      PreviousOK = !strncmp(tval->v.s,bval->v.s,i-1);
                    else
                      PreviousOK = TRUE;
                    if((ch==tval->v.s[i])&&PreviousOK) {
                      match=TRUE;
                      i++;
                    } else
                      tval=tval->next;
                  } while((!match)&&(tval!=copt->val));
                  mvwprintw(cwin->win,copt->y,copt->x,"%s",tval->v.s);
                  wclrtoeol(cwin->win);
                  if(tval==copt->val) {                /* error */
                    done=TRUE;
                    flash();
                  } else {                             /* set current guess */
                    bval=tval;           
                    wmove(cwin->win,copt->y,copt->x+i);
                  }
                }
              }
              copt->val=bval;
            }
          }
          mvwprintw(cwin->win,copt->y,copt->x,"%s",copt->val->v.s);
          wclrtoeol(cwin->win);
          wrefresh(cwin->win);
          break;
        }
      }
      noecho();
      continue;
    }
    
    wrefresh(cwin->win);
  } while(!done);
}

/*****************************************************************************/

/* shuts down curses */

PRIVATE void CloseWindow
(
 window *w      /* data structure */
)
{
  window *cwin=w;   /* current window */

  do {
    delwin(cwin->win);
    cwin=cwin->next;
  } while(cwin!=w);
}

/*****************************************************************************/

/* this function reads the data file and sets up the data structure */

PRIVATE window *ReadDataFile
(
 FILE *fptr   /* infile pointer */
)
{
  window *cwin=NULL;             /* current window */

  do {
    cwin=AddElement(cwin,sizeof(struct window));
  } while(GetWindow(cwin->prev,fptr));

  return(cwin);
}

/*****************************************************************************/

/* given the initial line, this function reads from fptr for the options
   and sets up the window, wptr.  It returns the first line of the NEXT
   window or NULL if EOF */

PRIVATE window *GetWindow
(
 window *wptr,                 /* pointer to window */
 FILE *fptr                    /* input file pointer */
)
{
  char *str=strbuff(fptr,NULL);     /* string to process */

                        /* confirm first line contains executable not option */
  if((str!=NULL)&&(str[0]=QUOTE)) {
    char *end=str+strlen(str);      /* last char in initial line */
    int stringlength;               /* length of current substring */
    
    stringlength=StringLength(str,QUOTE);
    wptr->exec=malloc(stringlength+1);
    str=StringCopy(wptr->exec,str,QUOTE,end-str);

    stringlength=StringLength(str,QUOTE);
    wptr->title=malloc(stringlength+1);
    str=StringCopy(wptr->title,str,QUOTE,end-str);
  }

  for(str=strbuff(fptr,NULL);(str!=NULL)&&(str[0]==MINUS);str=strbuff(fptr,NULL)) {
    char *end=str+strlen(str);    /* last char in line */
    int stringlength;             /* length of current substring */

    if(str!=NULL) {
      option *lptr;            /* current option */

      wptr->opt=AddElement(wptr->opt,sizeof(struct option));
      lptr=wptr->opt->prev;
      
      stringlength=StringLength(str,SPACE);                   /* option code */
      lptr->opt=malloc(stringlength+1);
      sscanf(str,"%s",lptr->opt);

      stringlength=StringLength(str,QUOTE);                    /* intro text */
      lptr->text=malloc(stringlength+1);
      str=StringCopy(lptr->text,str,QUOTE,end-str);

      do {
        char *t;                      /* ascii string to be parsed for value */
        
        lptr->val=AddElement(lptr->val,sizeof(struct val));
        stringlength=StringLength(str,SPACE);
        t=malloc(stringlength+1);
        str=StringCopy(t,str,SPACE,end-str);

        if(lptr->type==NONE)                                /* first element */
          if((isdigit(t[0]))||(isdigit(t[1])&&(t[0]==PERIOD)))  /* number */
            if(strchr(t,PERIOD)) 
              lptr->type = FLOAT;              /* float data */
            else
              lptr->type = INT;                /* int data */
          else
            lptr->type = STRING;               /* string data */

        switch(lptr->type) {
        case INT:
          lptr->val->prev->v.i = atoi(t);
          free(t);
          break;
        case FLOAT:
          lptr->val->prev->v.f = atof(t);
          free(t);
          break;
        case STRING:
          lptr->maxstr=MAX(lptr->maxstr,stringlength);
          lptr->val->prev->v.s=t;
          break;
        }
      } while(*str!=NUL);
    }
  }

  if(str==NULL)       /* nothing left to read */
    wptr=NULL;
  else
    strbuff(fptr,str);  /* push extra string onto buffer */

  return(wptr);
}

/*****************************************************************************/

PRIVATE void WriteScript
(
 char *fname,    /* output filename */
 window *w       /* double linked list of windows */
)
{
  window *cwin=w;                /* current window */
  FILE *fptr=fopen(fname,"w");
  
  if(fptr) {
    do {
      option *copt=cwin->opt;     /* pointer to current option */

      fprintf(fptr,"%s ",cwin->exec);  /* print program and fixed options */

      if(copt)                         /* variable options exist */
        do {                           /* print variable options */
          fprintf(fptr,"%s ",copt->opt);

          switch(copt->type) {
          case INT:
            fprintf(fptr,"%d ",copt->val->v.i);
            break;
          case FLOAT:
            fprintf(fptr,"%g ",copt->val->v.f);
            break;
          case CHAR:
            fprintf(fptr,"%c ",copt->val->v.c);
            break;
          case STRING:
            fprintf(fptr,"%s ",copt->val->v.s);
            break;
          }

          copt=copt->next;           /* next option */
        } while(copt!=cwin->opt);

      fprintf(fptr,"\n");

      cwin=cwin->next;         /* next window */
    } while(cwin!=w);
    fclose(fptr);
  } else
    perror("WriteScript: unable to open output file ");
}
